//-- "action" function

function ParseExpr(line)
{
  var reVar = /\$\{([^\}]+)\}/g;
  var varr;
  /*replacing ${expr}*/
  while((varr = reVar.exec(line)) != null)
  {
    SafeEval("__val="+varr[1]);
    line = line.replace(new RegExp("\\$\\{"+FixPattern(varr[1])+"\\}","g"),__val);
  }
  return line;
}

function ParseJMCVars(line)
{
  var reJMCVar = /\$\[(\w+)\]/g;
  var varr;
  /*replacing $[jmcvar]*/
  while((varr = reJMCVar.exec(line)) != null)
    line = line.replace(new RegExp("\\$\\["+varr[1]+"\\]","g"),Var(varr[1]));
  return line;
}

function DoParse(line)
{
  line = ParseJMCVars(line);
  if((line = ParseInput(line)) != null)
    jmc.Parse(ParseExpr(line));
}

function DoAntiSubst(line)
{
}

function DoSubst(line)
{
  jmc.Event = line;
}

function DoGag(line)
{
  jmc.Event = null;
}

//-- class Trigger

// options:
var ALIAS     = "a";
var NOFIX     = "n";
var GAG       = "g";
var FALLTHROW = "f";
var ALLOWJMC  = "j";
var DISABLED  = "-";
var PRIOR_DEFAULT = 500;

function FixPattern(pattern)
{
  pattern = pattern.replace(/([\.\/\\\(\(\)\[\]\*\+\?\{\}])/g,"\\$1");
  return pattern.replace(/%\d/g,"(.*)");
}

function FixMask(mask)
{
  mask = mask.replace(/([\^\$\|\.\/\\\(\(\)\[\]\+\?\{\}])/g,"\\$1");
  mask = mask.replace(/\*/g,".*");
  mask = mask.replace(/%\d/g,"\\(\\.\\*\\)");
  return mask;
}

function ConvertPatternToRegExp(pattern)
{
  pattern = FixPattern(pattern);
  return new RegExp(pattern);
}


function ConvertMaskToRegExp(mask)
{
  if(IsUndef(mask))
    return /.*/;
  else if(typeof(mask) != "string")
    return ConvertPatternToRegExp(mask.source);
  else
    mask = FixMask(mask);
  return new RegExp(mask);
}

function Trigger(regexp,action,options)
{
  var isshow = IsNull(regexp) || IsUndef(action);
  if(typeof(regexp)=="string" && !isshow)
      regexp = ConvertPatternToRegExp(regexp);
  this.regexp   = regexp;
  this.action   = action;
  this.prior    = PRIOR_DEFAULT;
  this.options  = new Options;

  var arr = /([nagfj-]*)(\d*)([nagfj-]*)(([,:;]\w+)*)?/.exec(options);
  if(arr != null)
  {
    if(arr[2] != "")
      this.prior = arr[2];
    this.options.Enable(arr[1]+arr[3]);
    this.groups = arr[4];
  }
  if(isshow)
    this.groups = "!show!";
}

Trigger.prototype.toString = function()
{
  return this.regexp + ' -> /' + this.action + '/'
             + " (" + this.prior + this.options + this.groups + ")";
}

function CompareTriggers(a,b)
{
  if(IsUndef(a) || a.options.Test(DISABLED))
    return 1;
  if(IsUndef(b) || b.options.Test(DISABLED))
    return -1;
  return a.prior - b.prior;
}

//-- class Triggers

function Triggers(func)
{
  this._func = func;
  this.array = new Array;
}

Triggers.prototype.Parse = function(str)
{
/*DBG  jmc.ShowMe("parsing:"+str);*/
  var parsed = false;
  for(var i=0;i<this.array.length;i++)
  {
    if((typeof(this.array[i]) != "object"))
      continue;
    var trigger = this.array[i];
    if(IsUndef(trigger) || trigger.options.Test(DISABLED))
      break;
    var arr = trigger.regexp.exec(str);
    if(arr!=null)
    {
/*DBG        jmc.ShowMe("matched with:"+trigger.regexp);*/
      parsed = true;
      if(!IsUndef(trigger.action))
      {
        var todo = trigger.action;
        if(typeof(todo) == "function")
          todo(arr);
        else
        {
          var left  = str.substr(0,arr.index);
          var right = str.substr(arr.lastIndex);
          /*replacing [$%]n by fixed arguments*/
          for(var no=0; no<arr.length; no++)
          {
            var arg = trigger.options.Test(NOFIX) ? arr[no] : FixArg(arr[no]);
            todo = todo.replace(new RegExp("[\\$%]"+no,"g"),arg);
          }
          /*replacing $< and $>*/
          todo = todo.replace(/[\$%]</g,FixArg(left));
          todo = todo.replace(/[\$%]>/g,FixArg(right));
          todo = ParseJMCVars(todo);
          if(right!="" && trigger.options.Test(ALIAS))
            todo += " " + right;
          this._func(todo);
        }
      }
      if(trigger.options.Test(GAG))
        jmc.Event = null;
      if(!trigger.options.Test(ALLOWJMC))
        jmc.DropEvent();
      if(!trigger.options.Test(FALLTHROW))
        break;
    }
  }
  return parsed;
}

Triggers.prototype.Add = function(trigger)
{
  if(trigger.groups == "!show!")
  {
    this.Show(trigger.regexp);
    return;
  }
  for(var i=0;i<this.array.length;i++)
    if(IsUndef(this.array[i]) || this.array[i].regexp.source==trigger.regexp.source)
      break;
  this.array[i] = trigger;
  this.Sort();
  jmc.ShowMe("%" + trigger);
}

Triggers.prototype.Sort = function(trigger)
{
  this.array.sort(CompareTriggers)
}

Triggers.prototype.Remove = function(mask)
{
  if(IsNull(mask))
    return;
  if(typeof(mask) == "string")
    mask = new RegExp("^"+FixMask(mask)+"$");

  for(var i=0;i<this.array.length;i++)
  {
    var trigger = this.array[i];
    if((typeof(trigger) != "object"))
      break;
    if(mask.test(trigger.regexp.source))
    {
      jmc.ShowMe("%-- " + trigger.regexp);
      delete this.array[i];
    }
  }
  this.Sort();
}

Triggers.prototype.Show = function (mask)
{
  mask = ConvertMaskToRegExp(mask);
  
  for(var i=0;i<this.array.length;i++)
  {
    var trigger = this.array[i];
    if((typeof(trigger) != "object"))
      continue;
    if(mask.test(trigger.regexp.source))
      jmc.ShowMe("%" + trigger);
  }
}

Triggers.prototype.EnableGroup = function (group)
{
  var re = new RegExp("\\b" + group + "\\b");
  for(var i=0;i<this.array.length;i++)
  {
    var trigger = this.array[i];
    if((typeof(trigger) != "object"))
      continue;
    if(re.test(trigger.groups))
      trigger.options.Disable(DISBALED);
  }
  this.Sort();
  jmc.ShowMe("%" + group + " group enabled.");
}

Triggers.prototype.DisableGroup = function (group)
{
  var re = new RegExp("\\b" + group + "\\b");
  for(var i=0;i<this.array.length;i++)
  {
    var trigger = this.array[i];
    if((typeof(trigger) != "object"))
      continue;
    if(!re.test(trigger.groups))
      trigger.options.Enable(DISBALED);
  }
  this.Sort();
  jmc.ShowMe("%" + group + " group disabled.");
}

AliasArray     = new Triggers(DoParse);
TriggerArray   = new Triggers(DoParse);
GagArray       = new Triggers(DoGag);
AntiSubstArray = new Triggers(DoAntiSubst);
SubstArray     = new Triggers(DoSubst);
HighlightArray = new Triggers(DoSubst);
